/*
 * hotlist.c: Routines to deal with the list of favourite VTX-pages
 *
 * Copyright (c) 1994-96 Martin Buck  <martin-2.buck@student.uni-ulm.de>
 * Read COPYING for more information
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/vtx.h>
#include "safe_malloc.h"
#include "tilde.h"
#include "misc.h"
#include "config.h"
#include "vtxtools.h"
#include "vtxqueue.h"
#include "toptext.h"
#include "hotlist.h"


#define TIMEOUT_DEFAULT 60
#define INTERLEAVE_DEFAULT 25
#define LOOKAHEAD_DEFAULT 2
#define EXPIRE_DEFAULT (7 * 24)
#define HOTLIST_FNAME "vtx-hotlist"
#define TOKEN_MAXLEN 100

typedef struct {
  char name[STATIONNAME_MAXLEN + 1], *inc_list, *ex_list;
  int count, *pages, page_timeout, top_interleave, lookahead, index_page, expire, freq, chan;
  tt_pageclass_t search_level;
  unsigned top_disable : 1;
  unsigned fast_disable : 1;
  unsigned readspool : 1;
  unsigned writespool : 1;
} hotlist_t;


static int hotlist_count, station_changed = TRUE;
static hotlist_t *hotlist, *current;
static char sp_list[800];



static int
parse_spool_list(char *list, const char *str, int newval) {
  char *tmp_str, *tokarg, *token, tmp_chr;
  unsigned int page1, page2;

  if (newval && list) {
    if (strpbrk(str, "0123456789,-")) {
      memset(list, FALSE, 800);			/* If include-list is empty, the default is to */
    } else {					/* spool all pages, otherwise only spool pages */
      memset(list, TRUE, 800);			/* in include-list */
    }
  }
  tokarg = tmp_str = sstrdup(str);
  while ((token = strtok(tokarg, ","))) {
    tokarg = NULL;
    if (sscanf(token, " %u - %u ", &page1, &page2) != 2) {
      if (sscanf(token, " %u %c", &page1, &tmp_chr) != 1) {
        page1 = 0;
      } else {
        page2 = page1;
      }
    }
    if (page1 < 100 || page1 > 899 || page2 < 100 || page2 > 899 || page1 > page2) {
      return FALSE;
    }
    if (list) {
      memset(list + page1 - 100, newval, page2 - page1 + 1);
    }
  }
  free(tmp_str);
  return TRUE;
}


static int
hotlist_load(const char *fname) {
  char line[256], token[TOKEN_MAXLEN + 1], tmpstr[PAGELIST_MAXLEN + 1], sep[2];
  int err, entry, stfound = 0, lcount = 0, tmpval, invalid, syntax;
  FILE *file;
  hotlist_t *newentry = NULL;

  if (!(file = fopen(fname, "r")))
    return -1;
  while (1) {
    if (!fgets(line, sizeof(line), file)) {
      if (feof(file))
        break;
      err = errno;
      fclose(file);
      errno = err;
      return -1;
    }
    lcount++;
    tmpstr[0] = '\0';
    invalid = syntax = FALSE;
    
    if (line[0] == '#' || sscanf(line, " %c ", token) <= 0) {	/* Skip comments or empty lines */
      continue;
    } else if (sscanf(line, " station %" STRINGIFY(STATIONNAME_MAXLEN) "[^\n\r] ", tmpstr) == 1) {
      stfound = 2;
      newentry = NULL;
      for (entry = 0; entry < hotlist_count; entry++) {
        if (!strcmp(tmpstr, hotlist[entry].name)) {
          newentry = hotlist + entry;
          break;
        }
      }
      if (newentry)
        continue;
      hotlist_count++;
      hotlist = current = srealloc(hotlist, hotlist_count * sizeof(hotlist_t));
      newentry = hotlist + hotlist_count - 1;
      *newentry = hotlist[0];
      strcpy(newentry->name, tmpstr);
      newentry->pages = smalloc(sizeof(int));
      newentry->count = 0;
      newentry->inc_list = sstrdup(hotlist[0].inc_list);
      newentry->ex_list = sstrdup(hotlist[0].ex_list);
    } else if (stfound != 2) {
      /* Ignore line if no station was given, but print warning only once */
      if (!stfound) {
        fprintf(stderr, "%s:%d: Warning: Missing station-name (ignoring entries).\n", fname,
            lcount);
        stfound = 1;
      }
    } else if (sscanf(line, " %" STRINGIFY(TOKEN_MAXLEN) "[^\1- =] %1[=] %"
        STRINGIFY(PAGELIST_MAXLEN) "[^\1-\37] ", token, sep, tmpstr) >= 2) {
      if (config_parse_int(token, tmpstr, "page_timeout", 10, 180, 10, &tmpval, &invalid)) {
        newentry->page_timeout = tmpval;
      } else if (config_parse_int(token, tmpstr, "top_interleave", -100, 100, 10, &tmpval,
          &invalid)) {
        newentry->top_interleave = tmpval;
      } else if (config_parse_int(token, tmpstr, "page_lookahead", 0, 5, 10, &tmpval, &invalid)) {
        newentry->lookahead = tmpval;
      } else if (config_parse_bool(token, tmpstr, "top_disable", &tmpval, &invalid)) {
        newentry->top_disable = tmpval;
      } else if (config_parse_bool(token, tmpstr, "fast_disable", &tmpval, &invalid)) {
        newentry->fast_disable = tmpval;
      } else if (config_parse_int(token, tmpstr, "index_page", 0x100, 0x8ff, 16, &tmpval,
          &invalid)) {
        if (!vtx_chkpgnum(tmpval, TRUE)) {
          invalid = TRUE;
        } else {
          newentry->index_page = tmpval;
        }
      } else if (config_parse_int(token, tmpstr, "expire", 0, 99 * 24 + 23, 10, &tmpval,
          &invalid)) {
        newentry->expire = tmpval;
      } else if (config_parse_int(token, tmpstr, "frequency", 0, INT_MAX, 10, &tmpval,
          &invalid)) {
        newentry->freq = tmpval;
      } else if (config_parse_int(token, tmpstr, "channel", 0, INT_MAX, 10, &tmpval,
          &invalid)) {
        newentry->chan = tmpval;
      } else if (config_parse_bool(token, tmpstr, "read_spooldir", &tmpval, &invalid)) {
        newentry->readspool = tmpval;
      } else if (config_parse_bool(token, tmpstr, "write_spooldir", &tmpval, &invalid)) {
        newentry->writespool = tmpval;
      } else if (!strcasecmp(token, "auto_search_pages")) {
        for (entry = TT_NONE; entry < TT_COUNT; entry++) {
          if (!strcasecmp(tmpstr, tt_pageclass_str[entry])) {
            newentry->search_level = entry;
            break;
          }
        }
        if (entry >= TT_COUNT) {
          invalid = TRUE;
        }
      } else if (!strcasecmp(token, "include_pages")) {
        if (parse_spool_list(NULL, tmpstr, TRUE)) {
          free(newentry->inc_list);
          newentry->inc_list = sstrdup(tmpstr);
        } else {
          invalid = TRUE;
        }
      } else if (!strcasecmp(token, "exclude_pages")) {
        if (parse_spool_list(NULL, tmpstr, FALSE)) {
          free(newentry->ex_list);
          newentry->ex_list = sstrdup(tmpstr);
        } else {
          invalid = TRUE;
        }
      } else {
        syntax = TRUE;
      }
    } else if (sscanf(line, " %x ", &tmpval) == 1) {
      if (!vtx_chkpgnum(tmpval, TRUE)) {
        invalid = TRUE;
      } else {
        newentry->pages = srealloc(newentry->pages, (newentry->count + 2) * sizeof(int));
        newentry->pages[newentry->count++] = tmpval;
      }
    } else {
      syntax = TRUE;
    }
    if (syntax) {
      fprintf(stderr, "%s:%d: Warning: Syntax error (ignoring line).\n", fname, lcount);
    } else if (invalid) {
      fprintf(stderr, "%s:%d: Warning: Invalid value (ignoring line).\n", fname, lcount);
    }
  }
  return fclose(file);
}


void
hotlist_init(void) {
  char *hl_fname;

  hotlist = current = smalloc(1 * sizeof(hotlist_t));
  strcpy(hotlist[0].name, "Unknown");
  hotlist[0].pages = smalloc(sizeof(int));
  hotlist[0].count = 0;
  hotlist[0].page_timeout = TIMEOUT_DEFAULT;
  hotlist[0].top_interleave = INTERLEAVE_DEFAULT;
  hotlist[0].lookahead = LOOKAHEAD_DEFAULT;
  hotlist[0].top_disable = FALSE;
  hotlist[0].fast_disable = TRUE;
  hotlist[0].index_page = 0x100;
  hotlist[0].search_level = TT_NORM;
  hotlist[0].readspool = hotlist[0].writespool = TRUE;
  hotlist[0].expire = EXPIRE_DEFAULT;
  hotlist[0].freq = 0;
  hotlist[0].chan = 0;
  hotlist[0].inc_list = sstrdup("");
  hotlist[0].ex_list = sstrdup("");
  hotlist_count = 1;

  memset(sp_list, TRUE, sizeof(sp_list));

  if (hotlist_load(VTX_LIBDIR "/" HOTLIST_FNAME) && errno != ENOENT) {
    fprintf(stderr, "%s: " VTX_LIBDIR "/" HOTLIST_FNAME ": %s\n", smalloc_progname,
        strerror(errno));
    exit(1);
  }

  hl_fname = sstrdup("~/." HOTLIST_FNAME);
  if (tilde_expand_realloc(&hl_fname) >= 0) {
    if (hotlist_load(hl_fname) && errno != ENOENT) {
      fprintf(stderr, "%s: %s: %s\n", smalloc_progname, hl_fname, strerror(errno));
      exit(1);
    }
  } else {
    fprintf(stderr, "%s: Warning: Can't load user-hotlist: No home directory found.\n",
        smalloc_progname);
  }
  free(hl_fname);
}


void
hotlist_exit(void) {
  int entry, page;
  const char *home;
  char *hl_fname, *hl_tmpfname;
  FILE *file;
  
  if (!(home = getenv("HOME"))) {
    fprintf(stderr, "%s: Warning: $HOME is unset -- can't save user-hotlist\n", smalloc_progname);
    return;
  }

  hl_fname = sstrdup("~/." HOTLIST_FNAME);
  if (tilde_expand_realloc(&hl_fname) >= 0) {
    hl_tmpfname = sstrdup(hl_fname);
    hl_tmpfname = sstrapp(hl_tmpfname, ".tmp");
    remove(hl_tmpfname);
    if ((file = fopen(hl_tmpfname, "w"))) {
      fprintf(file, "# This file was created by " VTXWINNAME ", Version " VTXVERSION "\n\n");
      for (entry = 0; entry < hotlist_count; entry++) {
        fprintf(file, "station %s\n"
            "page_timeout=%d\n"
            "top_interleave=%d\n"
            "page_lookahead=%d\n"
            "top_disable=%s\n"
            "fast_disable=%s\n"
            "index_page=%X\n"
            "auto_search_pages=%s\n"
            "read_spooldir=%s\n"
            "write_spooldir=%s\n"
            "expire=%d\n"
            "frequency=%d\n"
            "channel=%d\n",
            hotlist[entry].name,
            hotlist[entry].page_timeout,
            hotlist[entry].top_interleave,
            hotlist[entry].lookahead,
            (hotlist[entry].top_disable ? "true" : "false"),
            (hotlist[entry].fast_disable ? "true" : "false"),
            hotlist[entry].index_page,
            tt_pageclass_str[hotlist[entry].search_level],
            (hotlist[entry].readspool ? "true" : "false"),
            (hotlist[entry].writespool ? "true" : "false"),
            hotlist[entry].expire,
            hotlist[entry].freq,
            hotlist[entry].chan);
        if (entry) {
          /* Save exclude-list, include-list and hotlist only if station != Unknown */
          fprintf(file, "include_pages=%s\nexclude_pages=%s\n", hotlist[entry].inc_list,
              hotlist[entry].ex_list);
          for (page = 0; page < hotlist[entry].count; page++) {
            fprintf(file, "%3x\n", hotlist[entry].pages[page]);
          }
        }
        fprintf(file, "\n");
      }
    }
    if (!file || ferror(file) || fclose(file) || remove(hl_fname) < 0 ||
        rename(hl_tmpfname, hl_fname) < 0) {
      fprintf(stderr, "%s: Warning: Couldn't write %s: %s\n", smalloc_progname, hl_fname,
          strerror(errno));
      if (file) {
        fclose(file);
      }
    }
    remove(hl_tmpfname);
    free(hl_tmpfname);
  } else {
    fprintf(stderr, "%s: Warning: Can't save user-hotlist: No home directory found.\n",
        smalloc_progname);
  }
  free(hl_fname);
}


void
hotlist_add(int page, int pos) {
  current->pages = srealloc(current->pages, (current->count + 1) * sizeof(int));
  if (pos >= current->count)
    pos = current->count;
  memmove(current->pages + pos + 1, current->pages + pos, (current->count - pos) * sizeof(int));
  current->count++;
  current->pages[pos] = page;
}


int
hotlist_remove(int pos) {
  int retval;

  retval = current->pages[pos];
  memmove(current->pages + pos, current->pages + pos + 1,
      (current->count - pos - 1) * sizeof(int));
  /* Kluge to avoid getting a NULL pointer when alloc'ing 0 bytes: */
  current->pages = srealloc(current->pages,
      ((current->count - 1) > 0 ? (current->count - 1) : 1) * sizeof(int));
  current->count--;
  return retval;
}


int
hotlist_get(int pos) {
  return current->pages[pos];
}


int
hotlist_get_hotlist(int **list) {
  *list = current->pages;
  return current->count;
}


void
hotlist_set_hotlist(int *list, int count) {
  free(current->pages);
  if (count) {
    current->pages = smalloc(count * sizeof(int));
    memcpy(current->pages, list, count * sizeof(int));
  } else {
    current->pages = smalloc(sizeof(int));
  }
  current->count = count;
}


int
hotlist_get_count(int station) {
  return hotlist[station].count;
}


int
hotlist_get_stcount(void) {
  return hotlist_count;
}


void
hotlist_set_current(int station) {
  station_changed = TRUE;
  current = hotlist + station;
  parse_spool_list(sp_list, current->inc_list, TRUE);
  parse_spool_list(sp_list, current->ex_list, FALSE);
}


int
hotlist_get_current(void) {
  return current - hotlist;
}


void
hotlist_put_queue(void) {
  int page;
  
  for (page = 0; page < current->count; page++) {
    get_page(PRI_MED, current->pages[page], 0, 0, NULL, NULL);
    history_insert(current->pages[page], FALSE);
  }
}


char**
hotlist_mkstr(int *count) {
  static char **hotlist_str, *hotlist_pgstr;
  int entry;
  
  *count = current->count;
  if (hotlist_str) {
    hotlist_str = srealloc(hotlist_str, (*count + 1) * sizeof(char*));
    /* Kluge to avoid getting a NULL pointer when alloc'ing 0 bytes: */
    hotlist_pgstr = srealloc(hotlist_pgstr, (*count > 0 ? *count : 1) * 4 * sizeof(char));
  } else {
    hotlist_str = smalloc((*count + 1) * sizeof(char*));
    /* Kluge to avoid getting a NULL pointer when alloc'ing 0 bytes: */
    hotlist_pgstr = smalloc((*count > 0 ? *count : 1) * 4 * sizeof(char));
  }
  for (entry = 0; entry < *count; entry++) {
    hotlist_str[entry] = hotlist_pgstr + 4 * entry * sizeof(char);
    sprintf(hotlist_str[entry], "%3X", current->pages[entry]);
  }
  hotlist_str[*count] = NULL;
  return hotlist_str;
}


char*
hotlist_get_name(int entry) {
  return hotlist[entry].name;
}


static int
station_cmp(const byte_t *header, const byte_t *name) {
  int count, offset, len;
  const byte_t *header_chr, *name_chr;
  
  len = strlen(name);
  for (offset = 0; offset <= STATIONNAME_MAXLEN - len; offset++) {
    header_chr = header + offset;
    name_chr = name;
    for (count = 0; count <= STATIONNAME_MAXLEN - 1; count++) {
      if (!*name_chr)
        return TRUE;
      if (*header_chr && *header_chr != *name_chr)
        break;
      header_chr++;
      name_chr++;
    }
    if (!*name_chr) {
      return TRUE;
    }
  }
  return FALSE;
}


int
hotlist_search(const byte_t *vtx_name, int hint) {
  char name[STATIONNAME_MAXLEN];
  int pos, station, station_len = 0, station_num = 0;
  
  for (pos = 0; pos <= STATIONNAME_MAXLEN - 1; pos++) {
    name[pos] = vtx_name[pos];
    /* Don't match chars with parity-errors */
    if (!vtx_chkparity(&name[pos])) {
      name[pos] = '\1';
    }
    if (name[pos] < ' ') {
      name[pos] = ' ';
    }
    /* Ignore chars from national charsets when comparing station names, because the charset is
     * unknown until a page was found
     */
    switch (name[pos]) {
      case 0x23:
      case 0x24:
      case 0x40:
      case 0x5b:
      case 0x5c:
      case 0x5d:
      case 0x5e:
      case 0x5f:
      case 0x60:
      case 0x7b:
      case 0x7c:
      case 0x7d:
      case 0x7e:
        name[pos] = '\0';
      break;
    }
  }

  if (hint && station_cmp(name, hotlist[hint].name)) {
    return hint;
  }
  for (station = 1; station < hotlist_count; station++) {
    if (station != hint && station_cmp(name, hotlist[station].name) &&
        strlen(hotlist[station].name) > station_len) {
      station_num = station;
      station_len = strlen(hotlist[station].name);
    }
  }
  return station_num;
}


void
hotlist_new_name(const char *name, int station) {
  int old_station;

  /* When changing 'Unknown', create entry for new station and copy options from 'Unknown' */
  if (!station) {
    old_station = current - hotlist;
    hotlist = srealloc(hotlist, ++hotlist_count * sizeof(hotlist_t));
    hotlist[hotlist_count - 1] = hotlist[0];
    hotlist[0].pages = smalloc(sizeof(int));
    hotlist[0].count = 0;
    hotlist[0].inc_list = sstrdup("");
    hotlist[0].ex_list = sstrdup("");
    if (old_station) {
      current = hotlist + old_station;
    } else {
      current = hotlist + hotlist_count - 1;
    }
    strcpy(hotlist[hotlist_count - 1].name, name);
  } else {
    strcpy(hotlist[station].name, name);
  }
  station_changed = TRUE;
}


int
hotlist_station_changed(void) {
  int retval;
  
  retval = station_changed;
  station_changed = FALSE;
  return retval;
}


int
hotlist_get_timeout(void) {
  return current->page_timeout;
}


void
hotlist_set_timeout(int timeout) {
  current->page_timeout = timeout;
}


int
hotlist_get_interleave(void) {
  return current->top_interleave;
}


void
hotlist_set_interleave(int interleave) {
  current->top_interleave = interleave;
}


int
hotlist_get_lookahead(void) {
  return current->lookahead;
}


void
hotlist_set_lookahead(int lookahead) {
  current->lookahead = lookahead;
}


int
hotlist_get_topdisable(void) {
  return current->top_disable;
}


void
hotlist_set_topdisable(int disable) {
  current->top_disable = !!disable;
}


int
hotlist_get_fastdisable(void) {
  return current->fast_disable;
}


void
hotlist_set_fastdisable(int disable) {
  current->fast_disable = !!disable;
}


int
hotlist_get_index_page(void) {
  return current->index_page;
}


void
hotlist_set_index_page(int page) {
  current->index_page = page;
}


tt_pageclass_t
hotlist_get_searchlevel(void) {
  return current->search_level;
}


void
hotlist_set_searchlevel(tt_pageclass_t level) {
  current->search_level = level;
}


int
hotlist_get_spoolflag(int private, int read) {
/*FIXME*/
  return read ? current->readspool : current->writespool;
}


void
hotlist_set_spoolflag(int read, int flag) {
  if (read) {
    current->readspool = !!flag;
  } else {
    current->writespool = !!flag;
  }
}


int
hotlist_get_expire(void) {
  return current->expire;
}


void
hotlist_set_expire(int expire) {
  current->expire = expire;
}


const char*
hotlist_get_includelist(void) {
  return current->inc_list;
}


int
hotlist_set_includelist(const char *str) {
  if (!parse_spool_list(NULL, str, TRUE)) {
    return FALSE;
  }
  free(current->inc_list);
  current->inc_list = sstrdup(str);
  parse_spool_list(sp_list, current->inc_list, TRUE);
  parse_spool_list(sp_list, current->ex_list, FALSE);
  return TRUE;
}


const char*
hotlist_get_excludelist(void) {
  return current->ex_list;
}


int
hotlist_set_excludelist(const char *str) {
  if (!parse_spool_list(NULL, str, FALSE)) {
    return FALSE;
  }
  free(current->ex_list);
  current->ex_list = sstrdup(str);
  parse_spool_list(sp_list, current->inc_list, TRUE);
  parse_spool_list(sp_list, current->ex_list, FALSE);
  return TRUE;
}


int
hotlist_get_freq(void) {
  return current->freq;
}


void
hotlist_set_freq(int freq) {
  current->freq = freq;
}


int
hotlist_get_chan(void) {
  return current->chan;
}


void
hotlist_set_chan(int chan) {
  current->chan = chan;
}


int
hotlist_check_spool(int page) {
  page = vtx_hex2dec(page) - 100;
  return (page >= 0 && page <= 799 && sp_list[page]);
}


void
hotlist_copy(int station, int flags) {
  hotlist_t *old_current;
  /* Use initializers to make gcc happy */
  int timeout = 0, interleave = 0, topdisable = 0, fastdisable = 0, lookahead = 0;
  int index_page = 0, readspool = 0, writespool = 0, expire = 0, freq = 0, chan = 0;
  int hl_count = 0;
  int *hl_list = NULL;
  tt_pageclass_t searchlevel = TT_NONE;
  const char *includelist = NULL, *excludelist = NULL;

  if (flags & HL_HOTLIST) {
    hl_count = hotlist_get_hotlist(&hl_list);
  }
  if (flags & HL_TIMEOUT) {
    timeout = hotlist_get_timeout();
  }
  if (flags & HL_INTERLEAVE) {
    interleave = hotlist_get_interleave();
  }
  if (flags & HL_TOPDISABLE) {
    topdisable = hotlist_get_topdisable();
  }
  if (flags & HL_FASTDISABLE) {
    fastdisable = hotlist_get_fastdisable();
  }
  if (flags & HL_SEARCHLEVEL) {
    searchlevel = hotlist_get_searchlevel();
  }
  if (flags & HL_LOOKAHEAD) {
    lookahead = hotlist_get_lookahead();
  }
  if (flags & HL_INDEXPAGE) {
    index_page = hotlist_get_index_page();
  }
  if (flags & HL_READSPOOL) {
    readspool = hotlist_get_spoolflag(TRUE, TRUE);
  }
  if (flags & HL_WRITESPOOL) {
    writespool = hotlist_get_spoolflag(TRUE, FALSE);
  }
  if (flags & HL_EXINCLIST) {
    includelist = hotlist_get_includelist();
    excludelist = hotlist_get_excludelist();
  }
  if (flags & HL_EXPIRE) {
    expire = hotlist_get_expire();
  }
  if (flags & HL_FREQUENCY) {
    freq = hotlist_get_freq();
  }
  if (flags & HL_CHANNEL) {
    chan = hotlist_get_chan();
  }

  old_current = current;
  current = hotlist + station;
  if (flags & HL_HOTLIST) {
    hotlist_set_hotlist(hl_list, hl_count);
  }
  if (flags & HL_TIMEOUT) {
    hotlist_set_timeout(timeout);
  }
  if (flags & HL_INTERLEAVE) {
    hotlist_set_interleave(interleave);
  }
  if (flags & HL_TOPDISABLE) {
    hotlist_set_topdisable(topdisable);
  }
  if (flags & HL_FASTDISABLE) {
    hotlist_set_fastdisable(fastdisable);
  }
  if (flags & HL_SEARCHLEVEL) {
    hotlist_set_searchlevel(searchlevel);
  }
  if (flags & HL_LOOKAHEAD) {
    hotlist_set_lookahead(lookahead);
  }
  if (flags & HL_INDEXPAGE) {
    hotlist_set_index_page(index_page);
  }
  if (flags & HL_READSPOOL) {
    hotlist_set_spoolflag(TRUE, readspool);
  }
  if (flags & HL_WRITESPOOL) {
    hotlist_set_spoolflag(FALSE, writespool);
  }
  if (flags & HL_EXINCLIST) {
    hotlist_set_includelist(includelist);
    hotlist_set_excludelist(excludelist);
  }
  if (flags & HL_EXPIRE) {
    hotlist_set_expire(expire);
  }
  if (flags & HL_FREQUENCY) {
    hotlist_set_freq(freq);
  }
  if (flags & HL_CHANNEL) {
    hotlist_set_chan(chan);
  }
  current = old_current;
  parse_spool_list(sp_list, current->inc_list, TRUE);
  parse_spool_list(sp_list, current->ex_list, FALSE);
}
