/*
 * tuner.c: Routines to communicate with a tuner connected to a VTX-card's
 *          IC-Bus
 *
 * $Id: tuner.c,v 1.1 1996/09/22 22:38:20 mb Exp mb $
 *
 * Copyright (c) 1994-96 Martin Buck  <martin-2.buck@student.uni-ulm.de>
 * Read COPYING for more information
 *
 */

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/vtx.h>
#include "tuner.h"


char *tuner_device = "/dev/tuner";

static int tunfd = -1;



/* Open tuner-device. This doesn't reset the tuner, because then a
 * previously set frequency would be lost. So you should call tuner_reset()
 * yourself before accessing the tuner for the first time. This also checks,
 * if the driver-version is compatible (you have to pass the expected
 * major-version and the lowest possible minor- version); it also fills in
 * the tuner_info_t-struct pointed to by prg_info.
 * Return TUNOK if success, TUNEOPEN if device isn't ready, TUNERR if
 * I/O-error occured, TUNEVERSION if driver-version isn't compatible
 */
int
tuner_open(int major, int minor, tuner_info_t *prg_info) {
  tuner_info_t *info, tmp_info;
  
  if ((tunfd = open(tuner_device, O_RDONLY)) < 0)
    return TUNEOPEN;
  info = (prg_info != NULL ? prg_info : &tmp_info);
  if (ioctl(tunfd, TUNIOCGETINFO, info) < 0)
    return TUNERR;
  if (info->version_major != major || info->version_minor < minor) {
    while (close(tunfd) < 0 && errno == EINTR);
    tunfd = -1;
    return TUNEVERSION;
  }
  return TUNOK;
}


/* Close the tuner-device.
 */
void
tuner_close(void) {
  if (tunfd == -1)
    return;
  while (close(tunfd) < 0 && errno == EINTR);
  tunfd = -1;
}


/* Reset the tuner. This function may do different things depending on the
 * tuner used, but it always resets the frequency/channel. The only thing
 * you have to know is that you should call it at least once before using
 * the tuner-device.
 * Return TUNOK if success, TUNERR if I/O-error occured
 */
int
tuner_reset(void) {
  if (tunfd == -1)
    return TUNENOTOPEN;
  if (ioctl(tunfd, TUNIOCRESET) < 0)
    return TUNERR;
  return TUNOK;
}


/* Set tuner's frequency. The argument must be given in kHz and will be
 * rounded to the next value supported by the tuner.
 * Return TUNOK if success, TUNERR if I/O-error occured or TUNEINVAL if tuner
 * doesn't support setting the frequency directly or frequency is out of
 * range
 */
int
tuner_setfreq(int freq) {
  int err;
  
  if (tunfd == -1)
    return TUNENOTOPEN;
  if ((err = ioctl(tunfd, TUNIOCSETFREQ, (unsigned int)freq)) < 0) {
    return (errno == EINVAL) ? TUNEINVAL : TUNERR;
  }
  return TUNOK;
}


/* Get tuner's current frequency in kHz. This function might return 0 if the
 * tuner's frequency has never been set before by the
 * videotext-/tuner-driver, because in that case the current frequency is
 * unknown. The value returned might not be identical to the argument of the
 * last tuner_setfreq() call, because not all tuners support kHz-resolution.
 * Return TUNERR if I/O-error occured
 */
int
tuner_getfreq(void) {
  int freq;
  
  if (tunfd == -1)
    return TUNENOTOPEN;
  if (ioctl(tunfd, TUNIOCGETFREQ, &freq) < 0)
    return TUNERR;
  return freq;
}


/* Set tuner's channel. 
 * Return TUNOK if success, TUNERR if I/O-error occured or TUNEINVAL if
 * tuner doesn't support setting the channel directly or channel is out of
 * range
 */
int
tuner_setchan(int chan) {
  int err;
  
  if (tunfd == -1)
    return TUNENOTOPEN;
  if ((err = ioctl(tunfd, TUNIOCSETCHAN, (unsigned int)chan)) < 0) {
    return (errno == EINVAL) ? TUNEINVAL : TUNERR;
  }
  return TUNOK;
}


/* Get tuner's current channel. This function might return 0 if the tuner's
 * channel has never been set before by the videotext-/tuner-driver, because
 * in that case the current channel is unknown.
 * Return TUNERR if I/O-error occured
 */
int
tuner_getchan(void) {
  int chan;
  
  if (tunfd == -1)
    return TUNENOTOPEN;
  if (ioctl(tunfd, TUNIOCGETCHAN, &chan) < 0)
    return TUNERR;
  return chan;
}
